using System;
using System.Collections;
using System.Xml;

namespace stunnware.CRM.Fetch {
   /// <summary>
   /// The FetchCondition class specifies a condition inside a FetchXML query.
   /// </summary>
   public class FetchCondition {

      private string _attribute;
      private FetchConditionOperator _conditionOperator;
      private readonly ArrayList _values = new ArrayList();

      /// <summary>
      /// Default constructor
      /// </summary>
      public FetchCondition() {
      }

      /// <summary>
      /// Creates a new FetchCondition bound to the specified attribute.
      /// </summary>
      /// <param name="attribute">The attribute to use in the condition.</param>
      public FetchCondition(string attribute) {
         _attribute = attribute;
      }

      /// <summary>
      /// The attribute to use in the condition.
      /// </summary>
      public string Attribute {
         get {
            return _attribute;
         }

         set {
            _attribute = value;
         }
      }

      /// <summary>
      /// The condition operator.
      /// </summary>
      public FetchConditionOperator ConditionOperator {
         get {
            return _conditionOperator;
         }

         set {
            _conditionOperator = value;
         }
      }

      /// <summary>
      /// The value(s) used in the condition. Depending on the <see cref="ConditionOperator"/>, you need to specify more or less values.
      /// </summary>
      public ArrayList Values {
         get {
            return _values;
         }
      }

      /// <summary>
      /// Adds the definition of this class to an XML document.
      /// </summary>
      /// <param name="entityNode"></param>
      internal void AddToFetchDocument(XmlElement entityNode) {

         XmlElement conditionNode = FetchXmlDocument.AddElement(entityNode, "condition");
         FetchXmlDocument.AddAttribute(conditionNode, "attribute", Attribute);
         FetchXmlDocument.AddAttribute(conditionNode, "operator", ConditionOperatorText);

         if (Values != null) {

            bool useElement = 
               (this.ConditionOperator == FetchConditionOperator.In) ||
               (this.ConditionOperator == FetchConditionOperator.NotIn) ||
               (this.ConditionOperator == FetchConditionOperator.Between) ||
               (this.ConditionOperator == FetchConditionOperator.NotBetween);

            if (!useElement && (Values.Count == 1)) {
               FetchXmlDocument.AddAttribute(conditionNode, "value", Values[0].ToString());
            }

            else {
               foreach(object o in Values) {
                  XmlElement condition_value = FetchXmlDocument.AddElement(conditionNode, "value");
                  condition_value.InnerText = o.ToString();
               }
            }
         }
      }

      /// <summary>
      /// Parses the passed xml node and creates a new instance from it.
      /// </summary>
      /// <param name="node"></param>
      /// <returns></returns>
      public static FetchCondition FromXml(XmlNode node) {

         XmlAttribute attributeNode = node.Attributes["attribute"];
         XmlAttribute operatorNode = node.Attributes["operator"];

         FetchCondition condition = new FetchCondition(attributeNode.InnerText);
         condition.ConditionOperatorText = operatorNode.Value;

         XmlAttribute valueAttribute = node.Attributes["value"];

         if (valueAttribute != null) {
            condition.Values.Add(valueAttribute.Value);
         }

         else {
            XmlNodeList valueNodes = node.SelectNodes("value");

            foreach(XmlNode valueNode in valueNodes) {
               condition.Values.Add(valueNode.InnerText);
            }
         }

         return condition;
      }

      /// <summary>
      /// Converts the value specified in <see cref="ConditionOperator"/> to (get) or from (set) the appropriate operator used in FetchXml.
      /// </summary>
      private string ConditionOperatorText {
         get {
            switch(ConditionOperator) {

               case FetchConditionOperator.LowerThan: return "lt";
               case FetchConditionOperator.GreaterThan: return "gt";
               case FetchConditionOperator.LowerEqual: return "le";
               case FetchConditionOperator.GreaterEqual: return "ge";
               case FetchConditionOperator.Equal: return "eq";
               case FetchConditionOperator.NotEqual: return "ne";
               case FetchConditionOperator.Null: return "null";
               case FetchConditionOperator.NotNull: return "not-null";
               case FetchConditionOperator.In: return "in";
               case FetchConditionOperator.NotIn: return "not-in";
               case FetchConditionOperator.Between: return "between";
               case FetchConditionOperator.NotBetween: return "not-between";
               case FetchConditionOperator.Like: return "like";
               case FetchConditionOperator.NotLike: return "not-like";
               case FetchConditionOperator.Yesterday: return "yesterday";
               case FetchConditionOperator.Today: return "today";
               case FetchConditionOperator.Tomorrow: return "tomorrow";
               case FetchConditionOperator.NextSevenDays: return "next-seven-days";
               case FetchConditionOperator.LastSevenDays: return "last-seven-days";
               case FetchConditionOperator.NextWeek: return "next-week";
               case FetchConditionOperator.LastWeek: return "last-week";
               case FetchConditionOperator.ThisMonth: return "this-month"; 
               case FetchConditionOperator.LastMonth: return "last-month";
               case FetchConditionOperator.NextMonth: return "next-month";
               case FetchConditionOperator.On: return "on";
               case FetchConditionOperator.OnOrBefore: return "on-or-before";
               case FetchConditionOperator.OnOrAfter: return "on-or-after";
               case FetchConditionOperator.ThisYear: return "this-year";
               case FetchConditionOperator.LastYear: return "last-year"; 
               case FetchConditionOperator.NextYear: return "next-year"; 
               case FetchConditionOperator.EqualUserId: return "eq-userid"; 
               case FetchConditionOperator.NotEqualUserId: return "ne-userid"; 
               case FetchConditionOperator.EqualBusinessId: return "eq-businessid"; 
               case FetchConditionOperator.NotEqualBusinessId: return "ne-businessid"; 
               case FetchConditionOperator.LastXHours: return "last-x-hours"; 
               case FetchConditionOperator.NextXHours: return "next-x-hours"; 
               case FetchConditionOperator.LastXDays: return "last-x-days"; 
               case FetchConditionOperator.NextXDays: return "next-x-days";
               case FetchConditionOperator.LastXWeeks: return "last-x-weeks"; 
               case FetchConditionOperator.NextXWeeks: return "next-x-weeks"; 
               case FetchConditionOperator.LastXMonths: return "last-x-months"; 
               case FetchConditionOperator.NextXMonths: return "next-x-months"; 
               case FetchConditionOperator.LastXYears: return "last-x-years"; 
               case FetchConditionOperator.NextXYears: return "next-x-years";
               default: return "";
            }
         }

         set {

            if (value == null) {
               this.ConditionOperator = 0;
            }

            else {
               switch(value) {

                  case "lt": this.ConditionOperator = FetchConditionOperator.LowerThan; return;
                  case "gt": this.ConditionOperator = FetchConditionOperator.GreaterThan; return;
                  case "le": this.ConditionOperator = FetchConditionOperator.LowerEqual; return;
                  case "ge": this.ConditionOperator = FetchConditionOperator.GreaterEqual; return;
                  case "eq": this.ConditionOperator = FetchConditionOperator.Equal; return;
                  case "ne": this.ConditionOperator = FetchConditionOperator.NotEqual; return;
                  case "null": this.ConditionOperator = FetchConditionOperator.Null; return;
                  case "not-null": this.ConditionOperator = FetchConditionOperator.NotNull; return;
                  case "in": this.ConditionOperator = FetchConditionOperator.In; return;
                  case "not-in": this.ConditionOperator = FetchConditionOperator.NotIn; return;
                  case "between": this.ConditionOperator = FetchConditionOperator.Between; return;
                  case "not-between": this.ConditionOperator = FetchConditionOperator.NotBetween; return;
                  case "like": this.ConditionOperator = FetchConditionOperator.Like; return;
                  case "not-like": this.ConditionOperator = FetchConditionOperator.NotLike; return;
                  case "yesterday": this.ConditionOperator = FetchConditionOperator.Yesterday; return;
                  case "today": this.ConditionOperator = FetchConditionOperator.Today; return;
                  case "tomorrow": this.ConditionOperator = FetchConditionOperator.Tomorrow; return;
                  case "next-seven-days": this.ConditionOperator = FetchConditionOperator.NextSevenDays; return;
                  case "last-seven-days": this.ConditionOperator = FetchConditionOperator.LastSevenDays; return;
                  case "next-week": this.ConditionOperator = FetchConditionOperator.NextWeek; return;
                  case "last-week": this.ConditionOperator = FetchConditionOperator.LastWeek; return;
                  case "this-month": this.ConditionOperator = FetchConditionOperator.ThisMonth; return; 
                  case "last-month": this.ConditionOperator = FetchConditionOperator.LastMonth; return;
                  case "next-month": this.ConditionOperator = FetchConditionOperator.NextMonth; return;
                  case "on": this.ConditionOperator = FetchConditionOperator.On; return;
                  case "on-or-before": this.ConditionOperator = FetchConditionOperator.OnOrBefore; return;
                  case "on-or-after": this.ConditionOperator = FetchConditionOperator.OnOrAfter; return;
                  case "this-year": this.ConditionOperator = FetchConditionOperator.ThisYear; return;
                  case "last-year": this.ConditionOperator = FetchConditionOperator.LastYear; return; 
                  case "next-year": this.ConditionOperator = FetchConditionOperator.NextYear; return; 
                  case "eq-userid": this.ConditionOperator = FetchConditionOperator.EqualUserId; return; 
                  case "ne-userid": this.ConditionOperator = FetchConditionOperator.NotEqualUserId; return; 
                  case "eq-businessid": this.ConditionOperator = FetchConditionOperator.EqualBusinessId; return; 
                  case "ne-businessid": this.ConditionOperator = FetchConditionOperator.NotEqualBusinessId; return; 
                  case "last-x-hours": this.ConditionOperator = FetchConditionOperator.LastXHours; return; 
                  case "next-x-hours": this.ConditionOperator = FetchConditionOperator.NextXHours; return; 
                  case "last-x-days": this.ConditionOperator = FetchConditionOperator.LastXDays; return; 
                  case "next-x-days": this.ConditionOperator = FetchConditionOperator.NextXDays; return;
                  case "last-x-weeks": this.ConditionOperator = FetchConditionOperator.LastXWeeks; return; 
                  case "next-x-weeks": this.ConditionOperator = FetchConditionOperator.NextXWeeks; return; 
                  case "last-x-months": this.ConditionOperator = FetchConditionOperator.LastXMonths; return; 
                  case "next-x-months": this.ConditionOperator = FetchConditionOperator.NextXMonths; return; 
                  case "last-x-years": this.ConditionOperator = FetchConditionOperator.LastXYears; return; 
                  case "next-x-years": this.ConditionOperator = FetchConditionOperator.NextXYears; return;
               }
            }
         }
      }
   }

   /// <summary>
   /// Defines all possible condition operators
   /// </summary>
   public enum FetchConditionOperator {
      /// <summary>
      /// lt: lower than a specific value. Values must contain exactly 1 value.
      /// </summary>
      LowerThan,
      
      /// <summary>
      /// gt: greater than a specific value. Values must contain exactly 1 value.
      /// </summary>
      GreaterThan,

      /// <summary>
      /// le: lower than or equal to a specific value. Values must contain exactly 1 value.
      /// </summary>
      LowerEqual,

      /// <summary>
      /// ge: greater than or equal to a specific value. Values must contain exactly 1 value.
      /// </summary>      
      GreaterEqual,

      /// <summary>
      /// eq: equal to a specific value. Values must contain exactly 1 value.
      /// </summary>      
      Equal,

      /// <summary>
      /// ne: not equal to a specific value. Values must contain exactly 1 value.
      /// </summary>      
      NotEqual,

      /// <summary>
      /// null: checks for a null value. Values must be empty.
      /// </summary>      
      Null,

      /// <summary>
      /// not-null: checks for a non-null value. Values must be empty.
      /// </summary>      
      NotNull,

      /// <summary>
      /// in: checks if inside a list of values. Values must contain at least one value.
      /// </summary>
      In,

      /// <summary>
      /// not-in: checks if not inside a list of values. Values must contain at least one value.
      /// </summary>
      NotIn,

      /// <summary>
      /// between: checks if between two values. Values must contain two values, the first being the lower bound and the second
      /// being the upper bound.
      /// </summary>
      Between,

      /// <summary>
      /// not-between: checks if not between two values. Values must contain two values, the first being the lower bound and the second
      /// being the upper bound.
      /// </summary>      
      NotBetween,

      /// <summary>
      /// like: checks if like (similar to) a value. Values must contain exactly one value that can contain wildcards (_ and %).
      /// </summary>
      Like,

      /// <summary>
      /// not-like: checks if not like (similar to) a value. Values must contain exactly one value that can contain wildcards (_ and %).
      /// </summary>      
      NotLike,

      /// <summary>
      /// yesterday: checks if a date is on yesterday. Values must be empty.
      /// </summary>
      Yesterday,

      /// <summary>
      /// today: checks if a date is on today. Values must be empty.
      /// </summary>
      Today,

      /// <summary>
      /// tomorrow: checks if a date is on tomorrow. Values must be empty.
      /// </summary>
      Tomorrow,

      /// <summary>
      /// next-seven-days: checks if a date is in the next seven days. Values must be empty.
      /// </summary>
      NextSevenDays,

      /// <summary>
      /// last-seven-days: checks if a date is in the last seven days. Values must be empty.
      /// </summary>
      LastSevenDays,

      /// <summary>
      /// next-week: checks if a date is in the last week. Values must be empty.
      /// </summary>
      NextWeek,

      /// <summary>
      /// last-week: checks if a date is in the next week. Values must be empty.
      /// </summary>
      LastWeek,

      /// <summary>
      /// this-month: checks if a date is in this month. Values must be empty.
      /// </summary>
      ThisMonth,

      /// <summary>
      /// last-month: checks if a date is in the last month. Values must be empty.
      /// </summary> 
      LastMonth,

      /// <summary>
      /// next-month: checks if a date is in the next month. Values must be empty.
      /// </summary>
      NextMonth,

      /// <summary>
      /// on: checks if a date is on a specific date. Values must contain exactly one value.
      /// </summary>
      On,

      /// <summary>
      /// on-or-before: checks if a date is on or before a specific date. Values must contain exactly one value.
      /// </summary>
      OnOrBefore,

      /// <summary>
      /// on-or-after: checks if a date is on or fter a specific date. Values must contain exactly one value.
      /// </summary>      
      OnOrAfter,

      /// <summary>
      /// this-year: checks if a date is in this year. Values must be empty.
      /// </summary>
      ThisYear,

      /// <summary>
      /// last-year: checks if a date is in the last year. Values must be empty.
      /// </summary>
      LastYear,

      /// <summary>
      /// next-year: checks if a date is in the next year. Values must be empty.
      /// </summary>
      NextYear,

      /// <summary>
      /// eq-userid: checks if a systemuserid field equals the user id of the current user. Values must be empty.
      /// </summary>
      EqualUserId,

      /// <summary>
      /// ne-userid: checks if a systemuser or owner field is not equal to the user id of the current user. Values must be empty.
      /// </summary>
      NotEqualUserId,

      /// <summary>
      /// eq-businessid: checks if a businessunit or owner field equals the business unit id of the current user. Values must be empty.
      /// </summary>
      EqualBusinessId,

      /// <summary>
      /// ne-businessid: checks if a businessunit or owner field is not equal to the business unit id of the current user. Values must be empty.
      /// </summary>      
      NotEqualBusinessId,

      /// <summary>
      /// last-x-hours: checks if a date is in the last x hours. Values must contain exactly one value, specifying the X.
      /// </summary>
      LastXHours,

      /// <summary>
      /// next-x-hours: checks if a date is in the next x hours. Values must contain exactly one value, specifying the X.
      /// </summary>
      NextXHours,

      /// <summary>
      /// last-x-days: checks if a date is in the last x days. Values must contain exactly one value, specifying the X.
      /// </summary>
      LastXDays,

      /// <summary>
      /// next-x-days: checks if a date is in the next x days. Values must contain exactly one value, specifying the X.
      /// </summary>
      NextXDays,

      /// <summary>
      /// last-x-weeks: checks if a date is in the last x weeks. Values must contain exactly one value, specifying the X.
      /// </summary>
      LastXWeeks,

      /// <summary>
      /// next-x-weeks: checks if a date is in the next x weeks. Values must contain exactly one value, specifying the X.
      /// </summary>
      NextXWeeks,

      /// <summary>
      /// last-x-months: checks if a date is in the last x months. Values must contain exactly one value, specifying the X.
      /// </summary>
      LastXMonths,

      /// <summary>
      /// next-x-months: checks if a date is in the next x months. Values must contain exactly one value, specifying the X.
      /// </summary>
      NextXMonths,

      /// <summary>
      /// last-x-years: checks if a date is in the last x years. Values must contain exactly one value, specifying the X.
      /// </summary>
      LastXYears,

      /// <summary>
      /// next-x-years: checks if a date is in the next x years. Values must contain exactly one value, specifying the X.
      /// </summary>
      NextXYears
   }
}