/**************************************************************************************************
 * Tool:     FetchXML Library.
 * 
 * Purpose:  The FetchXML is a simple object model allowing you do define FetchXML queries the same
 *           way as you do with a QueryExpression. Though the SDK contains methods to convert a
 *           QueryExpression to a FetchXML query, there are some features in FetchXML that a
 *           QueryExpression cannot handle. This library supports the entire feature set and is
 *           very similar to use. If you've already used the QueryExpression class, it shouldn't
 *           be hard to understand. If you haven't used the QueryExpression, please visit the previous
 *           articles in the "Finding Data" series.
 * 
 *	Notes:    The serialization and deserialization usually will be done differently. The best way is
 *           to use XmlAttribute, XmlElement and XmlRoot attributes. If specified correctly, the
 *           XmlSerializer will do the job. As one of the main purposes of this library is to show how
 *           a FetchXML query is composed, I decided to do the serialization and deserialization manually, 
 *           using an XmlDocument. Though a XmlWriter should be faster, I thought that using an
 *           XmlDocument is easier to follow. Also this library is not meant to be used in an application
 *           that has to care about performance. It's mainly for instructional purposes.
 *  
 * Date:     07 Jan 2007
 * Author:   Michael Hhne, mailto:michael.hoehne@stunnware.com
 * DevEnv:   Microsoft Visual Studio 2003
 *  
 */
using System;
using System.Xml;

namespace stunnware.CRM.Fetch {
   /// <summary>
   /// The FetchStatement is the main class. It defines the whole FetchXML query.
   /// </summary>
   public class FetchStatement {

      private int _pageCount;
      private int _pageNumber = 1;
      private bool _aggregate;
      private bool _distinct;
      private string _pagingCookie; 
      private FetchEntity _fetchEntity;

      /// <summary>
      /// Default constructor.
      /// </summary>
      public FetchStatement() {
      }

      /// <summary>
      /// Creates a FetchStatement from the specified xml string.
      /// </summary>
      /// <param name="xml">A string containing an existing FetchXML query.</param>
      /// <returns>A FetchStatement representing the query.</returns>
      public static FetchStatement FromXml(string xml) {
         XmlDocument doc = new XmlDocument();
         doc.LoadXml(xml);
         return FromXml(doc);
      }

      /// <summary>
      /// Specifies the maximum number of records to retrieve. If set to 0, all records are returned.
      /// </summary>
      public int PageCount {
         get {
            return _pageCount;
         }

         set {
            _pageCount = (value > 0) ? value : 0;
         }
      }

      /// <summary>
      /// The page number when using <see cref="PageCount"/>.
      /// </summary>
      public int PageNumber {
         get {
            return _pageNumber;
         }

         set {
            _pageNumber = (value > 0) ? value : 0;
         }
      }

      /// <summary>
      /// If an aggregate is used in the query, this property must be set to true.
      /// </summary>
      public bool Aggregate {
         get {
            return _aggregate;
         }

         set {
            _aggregate = value;
         }
      }

      /// <summary>
      /// If true, only distinct results are returned. 
      /// </summary>
      public bool Distinct {
         get {
            return _distinct;
         }

         set {
            _distinct = value;
         }
      }

      /// <summary>
      /// The paging cookie must be specified to read the next page of results in a paged request.
      /// </summary>
      public string PagingCookie {
         get {
            return _pagingCookie;
         }

         set {
            _pagingCookie = value;
         }
      }

      /// <summary>
      /// The main entity.
      /// </summary>
      public FetchEntity Entity {
         get {
            return _fetchEntity;
         }

         set {
            _fetchEntity = value;
         }
      }

      /// <summary>
      /// Overrides the default implementation and returns the complete FetchXML query. Use this result as the input to the Fetch() method
      /// of the CRM wev service.
      /// </summary>
      /// <returns>The FetchXML query.</returns>
      public override string ToString() {
         
         FetchXmlDocument doc = new FetchXmlDocument();
         doc.AddRootAttribute("mapping", "logical");

         if (PageCount > 0) {
            doc.AddRootAttribute("count", PageCount);            
         }

         if (PageNumber > 1) {
            doc.AddRootAttribute("page", PageNumber);            
         }

         if (Aggregate) {
            doc.AddRootAttribute("aggregate", "true");
         }

         if (Distinct) {
            doc.AddRootAttribute("distinct", "true");
         }

         if (PagingCookie != null) {
            doc.AddRootAttribute("paging-cookie", PagingCookie);            
         }

         if (Entity != null) {
            Entity.AddToFetchDocument(doc);
         }

         return doc.ToString();
      }

      /// <summary>
      /// Parses the entire FetchXML query from an XML document.
      /// </summary>
      /// <param name="doc">The XML document containing the query.</param>
      /// <returns>The corresponding FetchStatement.</returns>
      public static FetchStatement FromXml(XmlDocument doc) {

         FetchStatement fetch = new FetchStatement();

         if (doc.DocumentElement == null) {
            throw new XmlException("The fetch element is missing.");
         }

         XmlAttribute attr = doc.DocumentElement.Attributes["count"];
         if (attr != null) {
            fetch.PageCount = int.Parse(attr.Value);
         }

         attr = doc.DocumentElement.Attributes["page"];
         if (attr != null) {
            fetch.PageNumber = int.Parse(attr.Value);
         }

         attr = doc.DocumentElement.Attributes["aggregate"];
         if (attr != null) {
            fetch.Aggregate = bool.Parse(attr.Value);
         }

         attr = doc.DocumentElement.Attributes["distinct"];
         if (attr != null) {
            fetch.Distinct = bool.Parse(attr.Value);
         }

         attr = doc.DocumentElement.Attributes["paging-cookie"];
         if (attr != null) {
            fetch.PagingCookie = attr.Value;
         }

         XmlNode entityNode = doc.DocumentElement.SelectSingleNode("entity");

         if (entityNode != null) {
            fetch.Entity = FetchEntity.FromXml(entityNode);
         }

         return fetch;
      }
   }
}

